#!/usr/bin/perl -w

package RPM::Source::Editor;
use strict;
use warnings;
use Carp;
use RPM::Source::SpecSection;
use RPM::Source::Macros;
use File::Basename;
use File::Temp qw/tempfile tempdir/;
use File::Path qw(make_path remove_tree);
use File::Spec;

our $VERSION=0.60;

sub new {
    my $class = shift;
    my $self = {
	CHANGELOG=> undef,
	WARN_NOTAPPLIED => 0,
	'-use_rpmbuild_-be' => 1,
#	'-fallback_hasher_tar' => 0,
#	'-readonly' => 0,
#	'-tmpdir' => undef,
#	VERBOSE => 0,
	CLEANUP => 1,
	PATCHDIR=> '.',
	@_
    };
    # hack against undefined VERBOSE
    $self->{VERBOSE} ||= 0;

    bless $self, $class;

    $self->{'-tmpdir'} = File::Spec->rel2abs( $self->{'-tmpdir'} ) if $self->{'-tmpdir'};

    if ($self->{SOURCERPM}) {
	if ($self->{SOURCERPM}!~/.src.rpm$/ && `file $self->{SOURCERPM}`!~/RPM/) {
	    die "$self->{SOURCERPM}: not a rpm file!\n";
	}
	$self->__mk_TMPDIR();
	$self->{SOURCEDIR}||="$self->{TMPDIR}/SOURCES";
	$self->{SPECDIR}||="$self->{TMPDIR}/SPECS";
	$self->{OUTPUTDIR}||="$self->{TMPDIR}/SRPMS";
	$self->rpm_custom('-i', $self->{SOURCERPM});
	$self->{SPECFILE}=(glob "$self->{TMPDIR}/SPECS/*")[0];
    } elsif ($self->{SPECFILE}) {
	# $self->{_SPECFILE_MODE_}=1;
    } else {
	die "$class: new: either SOURCERPM or SPECFILE is required."
    }
    my $curdir=`pwd`;
    chomp $curdir;
    foreach my $dir (qw/OUTPUTDIR SOURCEDIR SPECDIR TMPDIR/) {
	$self->{$dir}=$curdir.'/'.$self->{$dir} if $self->{$dir} and $self->{$dir}!~m!^/!;
    }
    $self->_set_speclist(load_file($self->{SPECFILE}));
    $self->{MACRO}=RPM::Source::Macros->new(SPECFILE=>$self->{SPECFILE}, VERBOSE=> $self->{VERBOSE});
    $self->_load_sources();
    #print STDERR "done _load_sources\n" if $self->{VERBOSE}>2;

    
    return $self;
}

sub macros {
    return $_[0]->{MACRO};
}

sub applied_off {
    my $self=shift;
    $self->{WARN_NOTAPPLIED}--;
}

sub applied_on {
    my $self=shift;
    $self->{WARN_NOTAPPLIED}++;
}

sub applied_block($$&) {
    my ($self, $name, $blockptr) = @_;
	$self->clear_applied();
	$self->applied_off();
	&$blockptr();
	$self->applied_on();
	$self->report_applied_flags($name);
}

sub clear_applied {
    my $self=shift;
    $self->{NOTAPPLIED_FLAG}=0;
    $self->{APPLIED_FLAG}=0;
}

sub is_not_applied {
    my $self=shift;
    return $self->{NOTAPPLIED_FLAG};
}

sub is_applied {
    my $self=shift;
    return $self->{APPLIED_FLAG};
}

sub set_applied {
    my $self=shift;
    $self->{APPLIED_FLAG}=1;
}

sub set_not_applied {
    my $self=shift;
    $self->{NOTAPPLIED_FLAG}=1;
}

sub report_applied_flags {
    my ($self, $name)=@_;
    if ($self->is_applied) {
	$self->_log_applied("$name is applied.");
    } else {
	$self->_warn_notapplied("$name is not applied.");
    }
}

sub _report_applied {
    my ($self, $is_applied, $name)=@_;
    if ($is_applied) {
	$self->_log_applied("$name is applied.");
    } else {
	$self->_warn_notapplied("$name is not applied.");
    }
}

sub _log_applied {
    my ($self)=@_;
    $self->{APPLIED_FLAG}=1;
}

sub _warn_notapplied {
    my $self=shift;
    $self->{NOTAPPLIED_FLAG}=1;
    my $wn=$self->{WARN_NOTAPPLIED};
    print STDERR "@_\n" if $wn and $wn>0;
}

sub __get_new_spec {
    my $self=shift;
    my $newspec=$self->{SPECFILE};
    unless ($self->{'-readonly'}) {
	my $newdir="$self->{TMPDIR}/NEW";
	mkdir $newdir;
	my $specname=basename($self->{SPECFILE});
	$newspec="$newdir/$specname";
	-e $newspec && unlink $newspec;
	&write_file($newspec, $self->get_spec());
    }
    return $newspec;
}

sub write_rpm {
    my $self=shift;
    my $newspec=$self->__get_new_spec();
    my @cmd=$self->__rpm_custom_arglist('-bs', '--nodeps', $newspec);
    return 1 if system(@cmd)==0;
    if (!$self->{'-fallback_hasher_tar'}) {
	Carp::croak ("@cmd failed: $?\nWriting src.rpm failed. May be some important rpm macros packages are missing\n");
    } else {
	warn ("Writing src.rpm failed. May be some important rpm macros packages are missing. Writing hasher tar.\n");
	$self->write_hasher_tar();
	return 0;
    }
}

sub write_hasher_tar {
    my $self=shift;
    my $specname=basename($self->{SPECFILE});
    my $newspec="$self->{SOURCEDIR}/$specname";
    -e $newspec && die "write_hasher_tar: Strange. $newspec already exists.";
    &write_file($newspec, $self->get_spec());
    make_path($self->{OUTPUTDIR});
    my $tarname=$self->macros()->macro_subst('%{name}-%{version}-%{release}.tar');
    my $tar_compress=''; # '--gzip' || '--bzip2'
    #&run("tar cf '$self->{OUTPUTDIR}/$tarname' -C '$self->{SOURCEDIR}' .");
    &run(qq[find '$self->{SOURCEDIR}' -maxdepth 1 -type f -printf '%f\\0' | LC_COLLATE=C sort -uz | xargs -r0 tar --create --file='$self->{OUTPUTDIR}/$tarname' --directory='$self->{SOURCEDIR}' --label='$specname' --owner=root --group=root --mode=u+w,go-w,go+rX ${tar_compress} -- ]);
    print STDERR "Written: $self->{OUTPUTDIR}/$tarname\n" if $self->{VERBOSE};
    unlink $newspec;
    return 1;
}

sub _kill_sections {
    my ($rpm)=@_;
    return unless ref $rpm->{SPECSECTIONS};
    map {%{$_}=(DESTROYED=>1)} @{$rpm->{SPECSECTIONS}};
}

sub _parse_speclist {
    my ($self, $specptr)=@_;
    my @outpart;
    my $out=[];
    my $TYPE='package';
    my $changelogmode;
    my $multilinedefine=0;
    foreach (@$specptr) {
	if ($multilinedefine) {
	    if (!/\\$/) {
		$multilinedefine=0;
	    }
	} elsif (/^\s*[%]define\s+[\w_\d]+(?:\(\))?.*\\$/) {
	    $multilinedefine=1;
	} elsif (/^(\s*)[%](\w+)\s*/) {
	    my $protospace=$1;
	    my $protosection=$2;
	    if ($protosection and $RPM::Source::SpecSection::SECTION{$protosection}) {
		print STDERR "entering _parse_speclist at \%$protosection\n" if $self->{VERBOSE}>2;
		print STDERR "WARNING: space padded section [$protosection]:$_" if $protospace and $self->{VERBOSE}>1;

		if ($protospace and $changelogmode) {
		    print STDERR "WARNING: ignored space padded section $protosection in changelog\n";
		} else {
		    $changelogmode=$protosection eq 'changelog';
		    push @outpart, RPM::Source::SpecSection->new(RPM=>$self, BODY=>$out,
			);#, TYPE=>$TYPE, PACKAGE=>$PACKAGE);
		    $out=[];
		}
	    }
	}
	push @$out, $_;
    }
    push @outpart, RPM::Source::SpecSection->new(RPM=>$self, BODY=>$out, TYPE => $TYPE);
    $self->_kill_sections();
    $self->set_sections(\@outpart);
}

# TODO: merge them? let _reset_speclist == _get_speclist?
sub _reset_speclist {
    my $self=shift;
    my @out= @{$self->_get_speclist()};
    $self->_set_speclist(\@out);
}

sub _get_speclist {
    my $self=shift;
    my @out;
    foreach (@{$self->{SPECSECTIONS}}) {
	push @out, map {split (/(?<=\n)/,$_)} @{$_->get_body()};
    }
    return \@out;
}

sub rename_package {
    my ($self, $newname) = @_;
    $self->{WARN_NOTAPPLIED}--;
    my $oldname=$self->macros()->{'name'};
    $self->macros()->{'name'}=$newname;
    $self->macros()->{'oldname'}=$oldname;
    for my $section ($self->get_sections()) {
	$section->subst(qr'%{name}','%{oldname}');
	$section->subst(qr'%name','%oldname');
	if ($section->get_type() eq 'package') {
	    $section->subst_if(qr'%{?oldname}?','%{name}',qr'^\s*Requires:\s*%{?oldname}?');
	}
    }
    my $mainsec=$self->get_section('package','');
    $mainsec->unshift_body('%define oldname '.$oldname."\n");
    $mainsec->set_tag('Name',$newname);
    my $prepsec=$self->get_section('prep');
    # bunch of quick hacks around setup; should be properly covered using arg parsing.
    $prepsec->subst(qr'%setup -q\n','%setup -q -n %{oldname}-%{version}'."\n");
    $prepsec->subst(qr'%setup\n','%setup -q -n %{oldname}-%{version}'."\n");
    $self->{WARN_NOTAPPLIED}++;
}


sub get_section {
    my ($self,$type,$pkg, $triggercond)=@_;
    $pkg='' if (!defined $pkg);
    foreach my $section (@{$self->{SPECSECTIONS}}) {
	# old behaviour; TODO: delete if 
	if ($triggercond) {
	    return $section if $section->get_type() eq $type and $self->are_package_names_equal($section->get_package(),$pkg) and $self->are_trigger_conditions_equal($section->get_trigger_condition(),$triggercond);
	} else {
	    return $section if $section->get_type() eq $type and $self->are_package_names_equal($section->get_package(),$pkg);
	}
    }
    return;
}

=head

del_section ($type,$pkg, $triggercond)

del_section ($section)

=cut

sub del_section {
    my ($self,$type,$pkg, $triggercond)=@_;

    if (ref $type && ref($type) eq 'RPM::Source::SpecSection') {
	my $section=$type;
	$type=$section->get_type();
	$pkg=$section->get_package();
	$triggercond=$section->get_trigger_condition();
    }

    $pkg='' if (!defined $pkg);
    Carp::croak "del_section: type not specified." unless $type;
    Carp::croak "main package section should not be deleted! use \$section->empty()!" if $pkg eq '' and $type eq 'package';
    my $found=0;
    my @new_sections;
    foreach my $section (@{$self->{SPECSECTIONS}}) {
	if (($section->get_type() ne $type) || !$self->are_package_names_equal($section->get_package(),$pkg) || !$self->are_trigger_conditions_equal($section->get_trigger_condition(),$triggercond)) {
	    push @new_sections, $section;
	} else {
	    $found=1;
	    $section->__kill_all_exept_macro_conditionals();
	    confess 'Oops: main section is to be deleted;' if @new_sections==0;
	    my $prev_section=$new_sections[$#new_sections];
	    $prev_section->__push_body_before_conditionals(@{$section->get_body()});
	    $prev_section->set_body(&__filter_empty_macro_if_endif($prev_section->get_body()));
	    
	}
    }
    $self->set_sections(\@new_sections);
    # done as $prev_section
    #$self->_reset_speclist();
    #$self->_set_speclist(&__filter_empty_macro_if_endif($self->_get_speclist()));
    if (!$found) {
	$self->_warn_notapplied("not deleted \%$type $pkg - not found");
    }

}

sub add_section {
    my ($self, $type, $pkg) =@_;
    $pkg ||= '';
    my $section=$self->get_section($type,$pkg);
    return $section if $section;
    $self->_raw_add_section($type, $pkg);
    return $self->get_section($type,$pkg);
}

my %scriptsection=map {$_=>1} qw/pre post preun postun pretrans posttrans triggerin triggerun triggerpostin triggerpostun/;

sub _raw_add_section {
    my ($self, $type, $pkg) =@_;
    $pkg ||= '';
    $pkg=$self->_section_name_to_canonical($pkg);
    my ($unshiftsec,$pushsec);
    if ($type eq 'package' || $type eq 'description') {
	$unshiftsec=$self->get_section('prep');
	$unshiftsec=$self->get_section('build') unless $unshiftsec;
	$unshiftsec=$self->get_section('install') unless $unshiftsec;
	warn 'STRANGE(add_section): no %prep,%build or %install' unless $unshiftsec;
    } elsif ($type eq 'prep') {
	foreach my $section (@{$self->{SPECSECTIONS}}) {
	    $pushsec=$section if $section->get_type() eq 'description';
	}
	warn 'STRANGE(add_section): no %description' unless $pushsec;
    } elsif ($type eq 'build') {
	$unshiftsec=$self->get_section('install');
	unless ($unshiftsec) {
	    foreach my $section (@{$self->{SPECSECTIONS}}) {
		$pushsec=$section if $section->get_type() eq 'description';
	    }
	}
	warn 'STRANGE(add_section): no %install or %description' unless $unshiftsec || $pushsec;
    } elsif ($scriptsection{$type}) {
	foreach my $section (@{$self->{SPECSECTIONS}}) {
	    $pushsec=$section if $scriptsection{$section->get_type()};
	}
    } elsif ($type eq 'files') {
	foreach my $section (@{$self->{SPECSECTIONS}}) {
	    $pushsec=$section if $section->get_type() eq 'files';
	}
	warn 'STRANGE(add_section): no %files' unless $pushsec;
    }
    $pushsec=$self->get_section('install') unless $pushsec;
    $pushsec=$self->get_section('build') unless $pushsec;
    $pushsec=$self->get_section('prep') unless $pushsec;
    if ($unshiftsec) {
	$unshiftsec->unshift_body_before_section("
\%$type $pkg

");
	print STDERR "DEBUG:_raw_add_section:unshiftsec\n" if $self->{VERBOSE}>2;
    } elsif ($pushsec) {
	$pushsec->push_body("
\%$type $pkg

");
	print STDERR "DEBUG:_raw_add_section:pushsec\n" if $self->{VERBOSE}>2;
    } else {
	Carp::cluck "can't place section \%$type $pkg :(";
    }
    $self->_reset_speclist();
}

#sub is_specfile_mode {
#    my $self=shift;
#    return $self->{_SPECFILE_MODE_};
#}

sub _set_speclist {
    my $self=shift;
    my $ptr=shift;
    die "not an array pointer!" if ! ref $ptr eq 'ARRAY';
    $self->_parse_speclist($ptr);
}

# just a dirty hack until %sec -args parsing will be done
sub _section_name_to_canonical {
    my $self=shift;
    my $sectionname=shift;
    if ($sectionname=~s/^\s*-n\s+(\S+)//) {
	my $name=$1;
	my $prefix=$self->macros()->macro_subst('%{name}');
	if ($name=~s/^$prefix-//) {
	    return $name.$sectionname;
	} elsif ($name eq $prefix) {
	    return $sectionname;
	}
	return '-n '.$name.$sectionname;
    }
    return $sectionname;
}

sub get_spec {
    my $self=shift;
    return join('',@{$self->_get_speclist()})
}

sub are_package_names_equal {
    my ($self, $name1, $name2)=@_;
    $name1=_package_name_strip_options($name1);
    $name2=_package_name_strip_options($name2);
    my $is_name1_n = substr($name1,0,2) eq '-n';
    my $is_name2_n = substr($name2,0,2) eq '-n';
    if ($is_name1_n or $is_name2_n) {
	$name1='-n %{name}-'.$name1 unless $is_name1_n;
	$name2='-n %{name}-'.$name2 unless $is_name2_n;
	# hack around comparison -n %{name} and ''
	$name1=~s/-$//;
	$name2=~s/-$//;
    }
    $name1=$self->macros()->macro_subst($name1);
    $name2=$self->macros()->macro_subst($name2);
    #print STDERR "are_package_names_equal: [$name1] [$name2]\n";
    # TODO: -n handler
    return $name1 eq $name2;
}

sub are_trigger_conditions_equal {
    my ($self, $cond1, $cond2)=@_;
    return 1 if !defined $cond1 && !defined $cond2;
    return 0 if !defined $cond1 || !defined $cond2;
    return 1 if $cond1 eq $cond2;
    return 0;
}

sub _package_name_strip_options {
    my $name=shift;
    $name=~s/^\s*//;
    $name=~s/\s*$//;
    while($name=~s/^-(?:l)\s+\w+\s*//) {};
    return $name;
}

sub disable_package {
    my $self=shift;
    my $name=shift;
    $name=~s/^\s*//;
    $name=~s/\s*$//;
    my $keyname=$name;
    $keyname=~s/^.+\s+//;
    $keyname=~s/-/_/g;
    unless ($keyname) {
	$keyname='main_package';
    }
    my $is_applied=0;
    unless ($name) {
	#disabling the main package
	my $section=$self->get_section('files','');
	if ($section) {
	    $section->unshift_body_before_section('%if_with '.$keyname."\n");
	    $section->push_body('%endif #'.$keyname."\n");
	    $is_applied = 1;
	}
    } else {
	#disabling a subpackage
	$self->{SPECSECTIONS}->[0]->unshift_body('%def_without '.$keyname."\n");
	for (my $i=0; $i<@{$self->{SPECSECTIONS}}; $i++) {
	    my $section = $self->{SPECSECTIONS}->[$i];
	    if ($self->are_package_names_equal($section->get_package(), $name)) {
		$section->unshift_body_before_section('%if_with '.$keyname."\n");
		$section->push_body('%endif #'.$keyname."\n");
		$is_applied = 1;
	    }
	}
    }
    if ($is_applied) {
	$self->_log_applied("disabled [$name]");
    } else {
	$self->_warn_notapplied("not disabled [$name] - not found");
    }
    return;
}

sub raw_rename_section {
    my $self=shift;
    my $oldname=shift;
    my $newname=shift;
    foreach my $section (@{$self->{SPECSECTIONS}}) {
	my $bodyref=$section->get_body();
	if (@$bodyref==0) {
	    #use Data::Dumper;
	    #warn "strange section ".Dumper($section);
	    warn "strange section with empty header: %".$section->get_type()." ".$section->get_package();
	    next;
	}
	$section->{PACKAGE}=$newname if ($bodyref->[0]=~s/(\s*\%\w+\s+)($oldname)\s*/$1$newname\n/);
    }
    return;
}

# TODO: perldoc
# note that add_section and del_section 
# TODO: should destroy (in fact, _set_speclist should destroy)
# objects returned by get_sections.
sub get_sections {
    my $self=shift;
    return @{$self->{SPECSECTIONS}};
}

sub set_sections {
    my ($self,$ptr)=@_;
    Carp::confess "argument to set_sections must be a pointer to array!" unless ref $ptr eq 'ARRAY';
    $self->{SPECSECTIONS}=$ptr;
}

sub _macros_refresh {
    my $self=shift;
    $self->{MACRO}=RPM::Source::Macros->new(SPECLIST=>$self->_get_speclist(), VERBOSE=> $self->{VERBOSE});
}

# TODO: remove?
sub set_default_changelog {
    my $self=shift;
    return $self->{CHANGELOG}=shift();
}

# TODO: check if the entry for the same version already exists
sub add_changelog {
    my $self=shift;
    my @opts=@_;
    # hack for add_changelog migration
    unshift @opts, 'ENTRY' if @opts==1;
    my $opt={@opts};
    my $entry=$opt->{ENTRY};
    $entry||=$self->{CHANGELOG};
    $entry ||='- ';

    # macros refresh (due to the usual raise in release).
    #$self->_macros_refresh();
    my $macros=$self->macros();

    my $packager=$opt->{PACKAGER};
    $packager||=$macros->macro_subst('%{packager}');
    $packager=`rpm --eval \%{packager}` if ! $packager || $packager eq '%{packager}';
    if (! $packager || $packager eq '%{packager}') {
	die "Can't guess packager: Configuration error: undefined packager";
    }

    my $date=$opt->{DATE};
    $date||=`LC_TIME=C date '+%a %b %d %Y'`;
    chomp $date;

    my $evr = $self->get_evr();
    my $qf="* $date $packager $evr\n";
    $entry.="\n" if (substr($entry,-2,1) ne "\n");
    $entry.="\n" if (substr($entry,-2,1) ne "\n");
    $self->get_section('changelog')->unshift_body($entry);
    $self->get_section('changelog')->unshift_body($qf);
}

sub get_evr {
    my $self=shift;
    my $mainsec = $self->get_section('package','');
    if ($self->{'-use_rpmbuild_-be'} && -e '/usr/bin/rpmbuild') {
	defined $self->{TMPDIR} || $self->__mk_TMPDIR();
	my $dumpspec=$self->{TMPDIR}.'/._______.spec';
	&write_array_to_file($dumpspec, $self->_get_speclist());
	if (system("rpmbuild -bE '$dumpspec' > '$dumpspec.0'")==0) {
	    my $specbE=RPM::Source::Editor->new(
		SPECFILE=> "$dumpspec.0",
		VERBOSE=> $self->{VERBOSE},
		);
	    $mainsec = $specbE->get_section('package','');
	}
	unlink $dumpspec;
	unlink "$dumpspec.0";
    }

    my $release = $mainsec->get_tag('Release');
    my $version = $mainsec->get_tag('Version');
    my $serial = $mainsec->get_tag('Serial');
    $serial = $mainsec->get_tag('Epoch') unless defined($serial) and $serial=~/^\d+$/;
    if (wantarray) {
	return $serial, $version, $release;
    } else {
	if (defined($serial) and $serial=~/^\d+$/) {
	    $serial.=':';
	} else {
	    $serial='';
	}
	return "${serial}${version}-${release}";
    }
}

sub copy_to_sources {
    my $self=shift;
    my $file=shift;
    my %param = (
	FILE => $file,
	NAME => undef,
	@_
	);

    # useful for specfile mode; rpm will be built in default location.
    unless ($self->{SOURCEDIR}) {
	$self->{SOURCEDIR}=$self->macros()->macro_subst(q!%_sourcedir!);
	make_path($self->{SOURCEDIR});
    }

    if ($param{NAME}) {
	$param{NAME} = basename($param{NAME});
    } else {
	$param{NAME} = basename($param{FILE});
    }
    $param{FILE} = "$self->{PATCHDIR}/$param{FILE}" unless (-e $param{FILE});
    run ('cp', '-f', $param{FILE}, $self->{SOURCEDIR}.'/'.$param{NAME});
}

sub add_patch {
    my $self=shift;
    my $name=shift;
    my %param = (
	NAME=> $name,
	STRIP => 1,
	NUMBER => 33,
	FILE => undef,
	@_
	);
    # find first unused patch position
    while ($self->{PATCHES}->[$param{NUMBER}]) {$param{NUMBER}++;}
    # register patch
    $self->{PATCHES}->[$param{NUMBER}]=$param{NAME};
    my $prep_section=$self->get_section('prep');
    my $main_section=$self->get_section('package','');
    my $patchline='%patch'.$param{NUMBER}.' -p'.$param{STRIP};
    $patchline.=' -b '.$param{BACKUP} if $param{BACKUP};
    # add patch to spec
    $prep_section->push_body($param{PREP_HEADER}) if $param{PREP_HEADER};
    $prep_section->push_body($patchline."\n");
    $prep_section->push_body($param{PREP_FOOTER}) if $param{PREP_FOOTER};
    $main_section->push_body($param{HEADER}) if $param{HEADER};
    $main_section->push_body('Patch'.$param{NUMBER}.': '.$param{NAME}."\n");
    $main_section->push_body($param{FOOTER}) if $param{FOOTER};
    $self->copy_to_sources($param{FILE} ? $param{FILE} : $param{NAME}, NAME => $param{NAME});
}

sub add_source {
    my $self=shift;
    my $name=shift;
    my %param = (
	NAME=> $name,
	NUMBER => 44,
	FILE => undef,
	@_
	);
    # find first unused patch position
    while ($self->{SOURCES}->[$param{NUMBER}]) {$param{NUMBER}++;}
    # register patch
    $self->{SOURCES}->[$param{NUMBER}]=$param{NAME};
    # add to spec
    my $main_section=$self->get_section('package','');
    $main_section->push_body($param{HEADER}) if $param{HEADER};
    $main_section->push_body('Source'.$param{NUMBER}.': '.$param{NAME}."\n");
    $main_section->push_body($param{FOOTER}) if $param{FOOTER};
    $self->copy_to_sources($param{FILE} ? $param{FILE} : $param{NAME}, NAME => $param{NAME});
    return $param{NUMBER};
}

sub rpm_custom {
    &run(&__rpm_custom_arglist);
}

sub __rpm_custom_arglist {
    my $self=shift;
    my @cmd =qq{rpm};
    push @cmd, '--define', "_specdir $self->{SPECDIR}" if $self->{SPECDIR};
    push @cmd, '--define', "_sourcedir $self->{SOURCEDIR}" if $self->{SOURCEDIR};
    push @cmd, '--define', "_srcrpmdir $self->{OUTPUTDIR}" if $self->{OUTPUTDIR};
    return(@cmd, @_);
}

sub DESTROY {
    my $self=shift;
    $self->cleanup() if $self->{CLEANUP};
}

sub cleanup {
    my $self=shift;
    #my $tmpdir=$self->{TMPDIR};
    #run("rm -rf $tmpdir") if $tmpdir;
    $self->__del_TMPDIR();
}

sub _load_sources {
    my $self=shift;
    print STDERR "entering _load_sources\n" if $self->{VERBOSE}>2;
    my $specfile = $self->{SPECFILE};
    my @source;
    my @patch;
    my @setup_n;
    my @is_setup;
    open (SPECFILE, $specfile) || die "can't open spec $specfile: $!";
    while (<SPECFILE>) {
	chomp;
	$source[$1 ? $1 : 0]=$3 if /^Source(\d*)\s*:\s*(.+\/)?([^\/]+)\s*$/;
	$patch[$1 ? $1 : 0]=$3 if /^Patch(\d*)\s*:\s*(.+\/)?([^\/]+)\s*$/;
	$setup_n[$1 ? $1 : 0] =$3 if /^[%]setup(\d*)\s*(.+)?-n\s+(\S+)/;
	$is_setup[$1 ? $1 : 0] =1 if /^[%]setup(\d*)/;
    }
    close (SPECFILE);

    my @rawsource=@source;
    my @rawpatch=@patch;
    foreach my $dep (grep {$_} @source, @patch) {
	print STDERR "loop3 $dep\n" if $self->{VERBOSE}>2;
	$self->macros()->macro_subst_inplace(\$dep);
    }
    $self->{SOURCES}=\@source;
    $self->{PATCHES}=\@patch;
    $self->{RAWSOURCES}=\@rawsource;
    $self->{RAWPATCHES}=\@rawpatch;
    $self->{SETUP_N}=\@setup_n;
    $self->{IS_SETUP}=\@is_setup;
}

#### utilities; ###########

sub run {
    my (@cmd)=@_;
    system(@cmd) && Carp::confess ("@cmd failed: $?");
}

sub load_file {
    my ($name)=@_;
    open FN, $name || confess "can't open $name: $!";
    my @result = <FN>;
    close FN, $name || die "can't open $name: $!";
    return \@result;
}

sub write_array_to_file {
    my ($name, $ptr)=@_;
    open FN, '>', $name || die "can't open $name: $!";
    print FN @$ptr;
    close FN, $name || die "can't open $name: $!";
}

sub write_file {
    my ($name, $str)=@_;
    open FN, '>', $name || die "can't open $name: $!";
    print FN $str;
    close FN, $name || die "can't open $name: $!";
}

##### private internals #########

sub __filter_empty_macro_if_endif {
    my ($inarrayptr)=@_;
    my @out;
    foreach my $line (@$inarrayptr) {
	push @out, $line;
	if ($line=~/^\s*\%endif/) {
	    my $position_of_if=__detect_empty_macro_if_endif(\@out);
	    $#out=$position_of_if-1 if $position_of_if >-1;
	}
    }
    return \@out;
}

# returns -1 on failure
# position of %if if found
sub __detect_empty_macro_if_endif {
    my $arrayptr=shift;
    my $is_empty=1;
    my $found_endif=0;
    my $found_else=0;
    my $pos=$#{$arrayptr};
    while ($pos>=0) {
	my $line=$arrayptr->[$pos];
	if ($line=~/^\s*\%endif/) {
	    return -1 if $found_endif;
	    return -1 if $found_else;
	    #print STDERR "found endif $line";
	    $found_endif=1;
	} elsif ($line=~/^\s*\%else/) {
	    return -1 if $found_else;
	    return -1 if ! $found_endif;
	    #print STDERR "found else $line";
	    $found_else=1;
	} elsif ($line=~/^\s*\%if/) {
	    #print STDERR "found pos=$pos\n";
	    return $pos;
	# do not kill comments
	#} elsif ($line!~/^\s*#/) {
	} elsif ($line!~/^\s*$/) {
	    return -1;
	}
	$pos--;
    }

}

sub __mk_TMPDIR {
    my $self=shift;
    if (!defined $self->{TMPDIR} || ! -d $self->{TMPDIR}) {
	if (defined $self->{'-tmpdir'}) {
	    make_path($self->{'-tmpdir'});
	    $self->{TMPDIR} = tempdir( "src-rpm-unpacked.XXXXXXXX", CLEANUP=>1, DIR => $self->{'-tmpdir'});
	} else {
	    $self->{TMPDIR} = tempdir( "src-rpm-unpacked.XXXXXXXX", CLEANUP=>1, TMPDIR => 1);

	}
	die "can't create temporary directory!" unless $self->{TMPDIR};
    } else {
	make_path($self->{TMPDIR});
    }
}

sub __del_TMPDIR {
    my $self=shift;
    if (defined $self->{TMPDIR} and -d $self->{TMPDIR}) {
	system 'chmod', '-Rf', 'u+rwX', $self->{TMPDIR};
	remove_tree ($self->{TMPDIR});
	undef $self->{TMPDIR};
    }
}

1;

=head1	NAME

RPM::Source::Editor -   for repocop

=head1	DESCRIPTION

TODO

=head1	AUTHOR

Written by Igor Vlasenko <viy@altlinux.org>.

=head1	COPYING

Copyright (c) 2008 Igor Vlasenko, ALT Linux Team.

This is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

=cut

