#!/usr/bin/perl -w

package RPM::Source::SpecSection;

use strict;
use warnings;
use Carp;
use RPM::Source::Macros;

our %SECTION=map {$_=>1} qw!
package
description
prep
build
install
check
clean
files
changelog
pre
post
preun
postun
pretrans
posttrans
triggerin
triggerun
triggerpostin
triggerpostun
!;

sub new {
    my $class = shift;
    my %flag;
    my $self= {
	TYPE=> 'package',
	PACKAGE=> '',
	TRIGGERCOND => undef,
	BODY=>[],
	FLAGS=> \%flag,
	@_
    };
    die "Oops! wrong BODY arg: should be an array ref" if ref $self->{BODY} ne 'ARRAY';
    #confess "Oops! empty BODY" if @{$self->{BODY}}==0;
    my $TYPE='package';
    my $PACKAGE='';
    my $header=$self->{BODY}->[0];
    $header=~s/\s*$//;
    if ($header=~s/^\s*\%(\w+)\s*// && $SECTION{$1}) {
	$TYPE=$1;
	while ($header=~s/^(\S+)\s*//) {
	    my $opt=$1;
	    if (substr ($opt,0,2) eq '--') {
		$self->{TRIGGERCOND}=$header;
		$header='';
	    } elsif (substr ($opt,0,1) eq '-') {
		$header=~s/^(\S+)\s*//;
		#print STDERR "flag: $opt=$1\n";
		$flag{$opt}=$1;
	    } else {
		die "Oops! strange section part: $opt $header (package=$PACKAGE)" if $PACKAGE;
		#print STDERR "package: [$opt]\n";
		$PACKAGE=$opt;
	    }
	}
	die "Oops! strange section: both -n $flag{'-n'} and package: $PACKAGE" if $PACKAGE and $flag{'-n'};
	$PACKAGE="-n $flag{'-n'}" if ($flag{'-n'});
    } 
    $self->{TYPE}=$TYPE;
    $self->{PACKAGE}=$PACKAGE;
    bless $self, $class;
#    return $self;
}

sub describe {
    my ($self,$noprint)=@_;
    &__check_self($self);
    my $description="SpecSection: type=$self->{TYPE}; package=$self->{PACKAGE}; body: ==>\n".join('',@{$self->{BODY}})."============== END BODY ====================\n";
    print $description unless $noprint;
    return $description;
}

sub get_flag {
    my ($self,$flag)=@_;
    &__check_self($self);
    return $self->{FLAGS}->{$flag};
}

sub get_type {
    my $self=shift;
    &__check_self($self);
    return $self->{TYPE};
}

sub get_trigger_condition {
    my $self=shift;
    &__check_self($self);
    return $self->{TRIGGERCOND};
}

sub rpm {
    my $self=shift;
    &__check_self($self);
    return $self->{RPM};
}

# TODO: get_package_name/get_package_header/?
# or get_name is better?
sub get_package {
    my $self=shift;
    &__check_self($self);
    my $pkgname=$self->{PACKAGE};
    $pkgname=~s/\s+-f\s+\S.*//;
    return $pkgname;
}

# TODO: get_package_name_expanded/get_package_header_expanded/?
sub get_package_expanded {
    my $self=shift;
    &__check_self($self);
    my $pkg= $self->get_package();
    my $retval;
    unless ($pkg) {
	$retval='%{name}';
    } elsif ($pkg=~s/-n\s*//) {
	$retval=$pkg;
    } else {
	$retval='%{name}-'.$pkg; 
    }
    $self->{RPM}->macros()->macro_subst_inplace(\$retval);
    return $retval;
}

sub get_body {
    my $self=shift;
    &__check_self($self);
    return $self->{BODY};
}

sub set_body {
    my $self=shift;
    my $ptr=shift;
    &__check_self($self);
    die "not an array pointer!" if ref $ptr ne 'ARRAY';
    foreach (@$ptr) {
	Carp::carp "set_body: got undefined value!" unless defined $_;
    }
    $self->{BODY}=$ptr;
}

sub unshift_body {
    my $self=shift;
    &__check_self($self);
    if ($self->is_head_section()) {
	unshift @{$self->{BODY}}, map {split (/(?<=\n)/,$_)} @_;
    } else {
	my $head = shift @{$self->{BODY}};
	unshift @{$self->{BODY}}, $head, map {split (/(?<=\n)/,$_)} @_;
    }
}

sub unshift_body_before_section {
    my $self=shift;
    &__check_self($self);
    unshift @{$self->{BODY}}, map {split (/(?<=\n)/,$_)} @_;
}

sub push_body {
    my $self=shift;
    &__check_self($self);
    my $bodyptr=$self->{BODY};
    # move pushed item(s) before %if* block of next section
    my @tail;
    my $i=$#{$bodyptr};
    while ($i>=0 and
	   # empty line or condition for the next section or comment;
	   # note: |\%else|\%endif is not skipped, as the line should not go into conditonal.
	   # The logic should be somehow rewritten using concept of line visibility.
	   $bodyptr->[$i]=~/^\s*(?:$|#|\%if)/ 
	) {
	unshift @tail, pop @{$bodyptr};
	$i--;
    }
    push @{$bodyptr}, map {split (/(?<=\n)/,$_)} @_;
    push @{$bodyptr}, @tail;
}

sub unshift_body_before {
    my $self=shift;
    my $toinsert = shift;
    my $pattern = shift;
    &__check_self($self);
    my $is_applied = 0;
    my @body=();
    my @tail=@{$self->{BODY}};
    # move pushed item(s) before %if* block of next section
    while (@tail>0 and $tail[0]!~/$pattern/ # condition for add
	) {
	push @body, shift @tail;
    }
    if (@tail) {
	push @body, split (/(?<=\n)/,$toinsert), @tail;
	$is_applied = 1;
    }
    $self->{BODY}=\@body;
    if ($is_applied) {
	$self->rpm->_log_applied("inserted [$toinsert]");
    } else {
	$self->rpm->_warn_notapplied("not inserted [$toinsert] - pattern [$pattern] not found");
    }
    return $is_applied;
}

sub unshift_body_after {
    my $self=shift;
    my $toinsert = shift;
    my $pattern = shift;
    &__check_self($self);
    my $is_applied = 0;
    my @body=();
    my @tail=@{$self->{BODY}};
    while (@tail>0 and $tail[0]!~/$pattern/ # condition for add
	) {
	push @body, shift @tail;
    }
    if (@tail) {
	push @body, shift @tail, split (/(?<=\n)/,$toinsert), @tail;
	$is_applied = 1;
    }
    $self->{BODY}=\@body;
    if ($is_applied) {
	$self->rpm->_log_applied("inserted [$toinsert]");
    } else {
	$self->rpm->_warn_notapplied("not inserted [$toinsert] - pattern [$pattern] not found");
    }
    return $is_applied;
}

sub push_body_after {
    my $self=shift;
    my $toinsert = shift;
    my $pattern = shift;
    &__check_self($self);
    my $bodyptr=$self->{BODY};
    my $is_applied = 0;
    my @tail;
    my $i=$#{$bodyptr};
    while ($i>=0 and $bodyptr->[$i]!~/$pattern/ # condition for add
	) {
	unshift @tail, pop @{$bodyptr};
	$i--;
    }
    if (@$bodyptr) {
	push @{$bodyptr}, split (/(?<=\n)/,$toinsert), @tail;
	$is_applied = 1;
    }
    if ($is_applied) {
	$self->rpm->_log_applied("inserted [$toinsert]");
    } else {
	$self->rpm->_warn_notapplied("not inserted [$toinsert] - pattern [$pattern] not found");
    }
    return $is_applied;
}

sub subst_after {
    my $self=shift;
    my $whatpattern=shift;
    my $replace=shift;
    my $afterpattern = shift;
    &__check_self($self);
    my $is_applied = 0;
    my $is_found_after=0;
    my @body=();
    my $count = 0;
    my $maxcount=1;
    my @tail=@{$self->{BODY}};
    while (@tail>0 and $tail[0]!~/$afterpattern/ # condition for add
	) {
	push @body, shift @tail;
    }
    $is_found_after = 1 if @tail;
    while (@tail>0 && $count <= $maxcount) {
	$_=shift @tail;
	#print STDERR "try:$_";
	if (s/$whatpattern/$replace/g) {
	    $is_applied = 1;
	    $count++;
	}
	push @body, $_;
    }
    push @body, @tail;
    $self->set_body(\@body);
    $is_found_after || $self->rpm->_warn_notapplied("pattern [$afterpattern] not found");
    $is_applied || $self->rpm->_warn_notapplied("not replaced [$whatpattern] after [$afterpattern]");
    $is_applied && $self->rpm->_log_applied("replaced [$whatpattern] after [$afterpattern]");
    return $is_applied;
}

sub empty {
    my $self=shift;
    &__check_self($self);
    # $#{$self->{BODY}}=0 if ($#{$self->{BODY}}>0);
    ## kill all except %if/%else/%endif
    my @newbody=($self->{BODY}->[0]);
    foreach (@{$self->{BODY}}[1..$#{$self->{BODY}}]) {
	push @newbody, $_ if /\s*\%(?:if|else|endif)/;
    }
    $self->{BODY}=\@newbody;
}

sub behead {
    my $self=shift;
    &__check_self($self);
    my @newbody=@{$self->{BODY}}[1..$#{$self->{BODY}}];
    $self->{BODY}=\@newbody;
}

sub is_empty {
    my $self=shift;
    &__check_self($self);
    return 1 if $#{$self->{BODY}}<1;
    return 0 if $self->{BODY}->[0]=~/^\%(post|pre|trigger).*\s+-p\s+\S/;
    foreach (@{$self->{BODY}}[1..$#{$self->{BODY}}]){
	#print "is_empty: [$_]" if ! /^\s*(#.*)?$/;
	return 0 if ! /^\s*(?:(?:#|\%if|\%else|\%endif).*)?$/;
    }
    return 1;
}

sub is_head_section {
    my $self=shift;
    &__check_self($self);
    return 1 if $self->{TYPE} eq 'package' && $self->{PACKAGE} eq '';
    return 0;
}

sub exclude {
    my $self=shift;
    my $pattern=shift;
    &__check_self($self);
    Carp::carp "exclude: got undefined argument" unless defined $pattern;
    my @out;
    my $is_applied = 0;
    foreach (@{$self->get_body()}) {
	unless (/$pattern/) {
	    push @out, $_;
	} else {
	    $is_applied++;
	}
    }
    $self->set_body(\@out);
    $self->rpm->_report_applied($is_applied,"exclude [$pattern]");
    return $is_applied;
}

sub exclude_multi {
    my ($self,@pattern)=@_;
    &__check_self($self);
    Carp::carp "exclude_multi: not enough arguments" unless scalar @pattern;
    my @out;
    my $is_applied = 0;
    my $i = 0;
    my @buffer;
    foreach my $line (@{$self->get_body()}) {
	if ($line=~/$pattern[$i]/) {
	    push @buffer, $line;
	    if (++$i==scalar @pattern) {
		$#buffer=-1;
		$i=0;
		$is_applied++;
	    }
	} else {
	    push @out, @buffer, $line;
	    $#buffer=-1;
	    $i=0;
	}
    }
    $self->set_body(\@out);
    $self->rpm->_report_applied($is_applied,"exclude_multi [@pattern]");
    return $is_applied;
}

sub subst {
    my $self=shift;
    my $pattern=shift;
    my $replace=shift;
    &__check_self($self);
    Carp::carp "subst: got undefined argument" unless defined $pattern and defined $replace;
    my $is_applied = 0;
    my @out;
    foreach (@{$self->get_body()}) {
	$is_applied++ if s/$pattern/$replace/g;
	push @out, $_;
    }
    $self->set_body(\@out);
    $self->rpm->_report_applied($is_applied,"subst [$pattern]");
    return $is_applied;
}

sub subst_if {
    my $self=shift;
    my $pattern=shift;
    my $replace=shift;
    my $ifpattern=shift;
    &__check_self($self);
    Carp::carp 'if_pattern is required' unless $ifpattern;
    my @out;
    my $is_applied = 0;
    foreach (@{$self->get_body()}) {
	if (/$ifpattern/) {
	    $is_applied++ if s/$pattern/$replace/g;
	}
	push @out, $_;
    }
    if ($is_applied) {
	$self->rpm->_log_applied("subst [$pattern]");
    } else {
	$self->rpm->_warn_notapplied("not subst [$pattern] if [$ifpattern]");
    }
    $self->set_body(\@out);
    return $is_applied;
}

sub replace_line {
    my $self=shift;
    my $stringwhat=shift;
    my $stringreplace=shift;
    &__check_self($self);
    my @out;
    foreach (@{$self->get_body()}) {
	$_ = $stringreplace if $_ eq $stringwhat;
	push @out, $_;
    }
    $self->set_body(\@out);
}



sub match {
    my $self=shift;
    my $pattern=shift;
    &__check_self($self);
    foreach (@{$self->get_body()}) {
	return 1 if /$pattern/;
    }
    return 0;
}

sub get_tag {
    my $self=shift;
    my $tag=shift;
    &__check_self($self);
    my %options = (
	RAW=>0,
	@_);
    foreach (@{$self->get_body()}) {
	if (/^$tag\s*:\s*(.+?)\s*$/i) {
	    #warn "tag=$tag raw =$1";
	    my $val=$1;
	    if (!$options{RAW}) {
		my $lctag=lc($tag);
		if ($val=~/\%\{?$lctag\}?/) {
		    # something like Release: alt1_%{release}1.7
		    # hack; assume %release is not overriden in subpackages
		    my $newval=$self->{RPM}->macros()->{$lctag};
		    $val=$newval if ($newval);
		} else {
		    $self->{RPM}->macros()->macro_subst_inplace(\$val);
		}
	    }
	    return $val;
	}
    }
    return;
}

sub set_tag {
    my ($self, $tag, $val)=@_;
    &__check_self($self);
    my $specptr=$self->get_body();
    my $found=0;
    for (my $i=0; $i<=$#{$specptr}; $i++) {
	if ($specptr->[$i]=~/^$tag:(\s*)/i) {
	    $specptr->[$i]="$tag:$1$val\n";
	    $found=1;
	    last;
	}
    }
    $self->unshift_body($tag.': '.$val."\n") unless $found;
    my $lctag=lc($tag);
    $self->{RPM}->macros()->{$lctag}=$self->{RPM}->macros()->macro_subst($val);
    return $self;
}

sub clear_tag {
    my ($self, $tag)=@_;
    &__check_self($self);
    my $specptr=$self->get_body();
    for (my $i=0; $i<=$#{$specptr}; $i++) {
	if ($specptr->[$i]=~/^$tag:(\s*)/i) {
	    $specptr->[$i]='#'.$specptr->[$i];
	    #last;
	}
    }
    my $lctag=lc($tag);
    undef $self->{RPM}->macros()->{$lctag};
    return $self;
}

### private internals ###
sub __check_self {
    my $self=shift;
    Carp::confess ("attempt to manipulate with dead section!") if $self->{DESTROYED};
}

sub __kill_all_exept_macro_conditionals {
    my ($self, $tag, $val)=@_;
    &__check_self($self);
    my @newbody=grep {/^\s*\%if/ || /^\s*\%else/ || /^\s*\%endif/} map {split /(?<=\n)/} @{$self->get_body()};
    $self->set_body(\@newbody);
}

sub __push_body_before_conditionals {
    my $self=shift;
    &__check_self($self);
    my $bodyptr=$self->{BODY};
    # move pushed item(s) before %if* block of next section
    my @tail;
    my $i=$#{$bodyptr};
    while ($i>=0 and
	       $bodyptr->[$i]=~/^\s*$/ # empty line
	) {
	unshift @tail, pop @{$bodyptr};
	$i--;
    }
    push @{$bodyptr}, @_, @tail;
}


1;
