#!/usr/bin/perl -w

package RPM::Source::Transform;

use strict;
use warnings;
use Carp;
use Getopt::Long;
use File::Basename;
use RPM qw/vercmp evrcmp/;
use RPM::Source::Editor;
use File::Temp qw/tempfile tempdir/;
#Getopt::Long::Configure ("bundling");

our ($inrpm,$parentrpm,$inspec,$parentspec,$parent,$inplace,@hooks,
     $changelog,$changelog_packager,$changelog_entry,$changelog_date,
     $packager,$default_packager_policy,$release,$epoch,$serial,
     $rename,$output_hasher_tar,$opt_tmpdir);
# transform mode: parent is igonored at release increment,
# changelog is from origin, not from parent;
our $force_transform_mode;

our $verbose=1;
our $fallback_hasher_tar=1;
our $defaulthook=1;
our $cleanup=1;
our $outputdir='.';
our $patchdir=dirname($0)."/patches";
our $nextrel='nmu';
our $defaultchangelog='- transformed by RPM::Source::Editor utils for unknown reason.';

our %NEXTREL_METHOD=(
    'incr' => \&nextrel_incr,
    'incrfirst' => \&nextrel_incrfirst,
    'incrsecond' => \&nextrel_incrsecond,
    'qa' => \&nextrel_qa,
    'nmu' => \&nextrel_nmuappend,
    'nmuappend' => \&nextrel_nmuappend,
    'nmuadd' => \&nextrel_nmuadd,
    'none' => \&nextrel_none,
    );

our %INITREL_METHOD=(
    );

sub new {
    my $class = shift;
    my $self = {
	'-specargs' => [],
	@_
    };
    bless $self, $class;
    return $self;
}

sub usage () {
    my ($self)=@_;
    print "options:
	-h, --help
	-v, --verbose (can be specified multiple times)
input options:
	[--rpm] /path/to/src.rpm
	[--spec] /path/to/spec
parent options:
	--parent[-rpm] /path/to/parent/src.rpm
	--parent[-spec] /path/to/parent/spec
Supplying parent spec/srpm is optional. If parent spec is supplied,
parent changelog, Serial/Epoch and possibly Packager tag will be merged.
output options:
	--outdir /path/to/output_dir
	-i|--inplace replace existing spec file (use together with --spec)
processing options:
	--hashertar write hasher-ready tar file instead of src.rpm file.
	--hook /path/to/hook to be called on resulting spec file.
	--hook option can be specified multiple times.
	--no-defaulthook do not look for default hook 'name.pl'
	--tag-packager-replace 'new value' set Packager: tag to a new value.
	  special values are: 
	    auto - use rpm's \%{packager} value
	    none - clear tag
	    parent - use parent's value of Packager:
	--tag-packager-default 'new value' - if Packager: tag is not defined 
	  set it to a new value. This option allows comma separated list
	  of multiple sources, for example, 'parent,A. R. <ar\@b.org>'
	-c|--changelog '- message'
	--rename 'new name' 
	--serial 'new value'
	--epoch 'new value'
	--release 'release' - if we need to adjust the package's release
	--nextrel|--next-release-policy <policy>
	  known next release policies: 
	".join(', ',sort(keys %NEXTREL_METHOD))."
	  default policy is $nextrel.
misc options:
	--tmpdir /path/to/alternate/tmp/dir
";
}

our @DEFAULT_OPTIONS=(
    "c|changelog=s"   => \$changelog,
    "changelog-entry=s"   => \$changelog_entry,
    "changelog-packager=s"   => \$changelog_packager,
    "changelog-date=s"   => \$changelog_date,
    "cleanup!"   => \$cleanup,
    "defaulthook!"   => \$defaulthook,
    "epoch=s"   => \$epoch,
    'hashertar'   => \$output_hasher_tar,
    "hook=s"   => \@hooks,
    "i|inplace"   => \$inplace,
    "nextrel|next-release-policy=s" => \$nextrel,
    "outdir=s"   => \$outputdir,
    "packager=s"   => \$packager,
    "parent=s"   => \$parent,
    "parent-spec=s"   => \$parentspec,
    "parent-rpm=s"   => \$parentrpm,
    "release=s"   => \$release,
    "rename=s"   => \$rename,
    "rpm=s" => \$inrpm,
    "serial=s"   => \$serial,
    "spec=s" => \$inspec,
    "tag-packager-default=s" => \$default_packager_policy,
    "tag-packager-replace=s" => \$packager,
    "tmpdir=s" => \$opt_tmpdir,
    "v|verbose+"  => \$verbose,
);

sub verify_options {
    my ($self)=@_;
    if (not $inrpm and not $inspec and @ARGV) {
	if ($ARGV[0]=~/\.src\.rpm$/) {
	    $inrpm = shift @ARGV;
	} elsif ($ARGV[0]=~/\.spec$/) {
	    $inspec = shift @ARGV;
	} else {
	    print STDERR "WARNING: $ARGV[0] does not look like src.rpm or spec.\n";
	} 
    }
    if (not $inrpm and not $inspec) {
	print STDERR "ERROR: please, specify [ --rpm | --spec ] /path/to/file argument.\n";
	$self->usage();
	exit (4);
    }
    if (not $parentrpm and not $parentspec and $parent) {
	if ($parent=~/\.src\.rpm$/) {
	    $parentrpm = $parent;
	} elsif ($parent=~/\.spec$/) {
	    $parentspec = $parent;
	} else {
	    print STDERR "WARNING: $parent does not look like src.rpm or spec.\n";
	} 
    }
}

sub process_options {
    my ($self)=@_;
    my $result = GetOptions (@DEFAULT_OPTIONS);
    $self->verify_options();
}

sub callback_hooks {
    my ($self)=@_;
}

sub transform {
    my ($self)=@_;

    my ($spec,$parent);

    my @tmpdir_opts;
    unless ($cleanup) {
    # all those tmp_dirname hacks are just to have at hand the RPM::Source::Editor guts on crash.
	my $tmp_dirname=basename($RPM::Source::Transform::inrpm);
	$tmp_dirname=~s!.src.rpm!!;
	my $tmp_pwd= `pwd`;
	chomp $tmp_pwd;
	$tmp_dirname="$tmp_pwd/$tmp_dirname.$$";
	push @tmpdir_opts, 	TMPDIR => $tmp_dirname;
    }
    $spec=RPM::Source::Editor->new(
	SOURCERPM => $inrpm, 
	SPECFILE=> $inspec, 
	OUTPUTDIR=> $outputdir, 
	PATCHDIR=>$patchdir, 
	VERBOSE=> $verbose,
	CLEANUP=> $cleanup,
	CHANGELOG=>$defaultchangelog,
	'-tmpdir' => $opt_tmpdir,
	'-fallback_hasher_tar' => $fallback_hasher_tar,
	@tmpdir_opts,
	@{$self->{'-specargs'}}
	);
    die "FATAL: failed to load the origin ".($inrpm ? $inrpm : $inspec).".\n" unless $spec;

    if ($parentrpm or $parentspec) {
	$parent=RPM::Source::Editor->new(
	    SOURCERPM => $parentrpm,
	    SPECFILE => $parentspec,
	    '-tmpdir' => $opt_tmpdir,
	    VERBOSE=> $verbose);
	die "FATAL: failed to load the parent ".($parentrpm ? $parentrpm : $parentspec).".\n" unless $parent;
    }
    
    # our -- hack ?
    our $jpptoolsdir=dirname($0);
    # for cross-references in hooks
    push @INC, $jpptoolsdir."/hooks", "./hooks", $jpptoolsdir."/common", "./common";
    
    if (0==@hooks or $defaulthook) {
	# looking for default hook
	my $name = $spec->get_section('package','')->get_tag('Name');
	$name=$rename if $rename;
	foreach my $defhook("$jpptoolsdir/hooks/$name.pl","./hooks/$name.pl") {
	    # macro_expansion should be here
	    #print "looking for default hook $defhook\n";
	    push @hooks, $defhook if (-e $defhook);
	}
    }

    our @SPECHOOKS;
    our @PREHOOKS;
    if (@hooks) {
	foreach my $hook (@hooks) {
	    eval {
		require $hook;
	    };
	    if ($@ && $hook!~/\.pl$/) {
		eval {
		    require $hook.'.pl';
		};
	    }
	    die "can't load hook $hook: $@\n" if $@;
	    die "script $hook does not have a valid hook!\n" unless @PREHOOKS or @SPECHOOKS;
	    if (@PREHOOKS) {
		$spec->{WARN_NOTAPPLIED}++;
		map {&$_($spec,$parent)} @PREHOOKS;
		$spec->{WARN_NOTAPPLIED}--;
	    }
	}
    }

    # to allow hooks write to changelog
    $spec->set_default_changelog($changelog) if ($changelog);
    $spec->rename_package($rename) if $rename;
    $self->callback_hooks($spec,$parent);

    if (@hooks) {
	if (@SPECHOOKS) {
	    $spec->{WARN_NOTAPPLIED}++;
	    map {&$_($spec,$parent)} @SPECHOOKS;
	    $spec->{WARN_NOTAPPLIED}--;
	}
    }

    &__set_serial	($spec,$parent,$serial,$epoch);
    &__increment_release($spec,$parent,$nextrel,$release,$force_transform_mode);
    &__set_packager	($spec,$parent,$default_packager_policy,$packager);
    &__set_changelog	($spec,$parent,$changelog_entry,$changelog_packager,$changelog_date,$force_transform_mode);

    $parent->cleanup() if $parent;

    my $write_rpm_ok=1;
    if ($inrpm) {
	print STDERR $spec->get_spec() if $verbose>3;
	if ($output_hasher_tar) {
	    $spec->write_hasher_tar();
	} else {
	    $write_rpm_ok=$spec->write_rpm();
	}
    } else { # spec
	if (!$inplace) {
	    print $spec->get_spec();
	} else {
	    system ('mv',$inspec,$inspec.'~');
	    open(my $fh, '>', $inspec) or die $!;
	    print $fh $spec->get_spec();
	    close ($fh);
	}
    }
    $spec->cleanup() if $cleanup || $write_rpm_ok;
}

sub __set_changelog {
    my ($spec,$parent,$changelog_entry,$changelog_packager,$changelog_date,$force_transform_mode)=@_;
    if ($parent && !$force_transform_mode) {
	$spec->get_section('changelog')->set_body($parent->get_section('changelog')->get_body());
    }
    # old hack to re-split lines against broken add_changelog entries
    $spec->get_section('changelog')->set_body($spec->get_section('changelog')->get_body());

    my @args;
    push @args, 'ENTRY' => $changelog if $changelog_entry;
    push @args, 'PACKAGER' => $changelog_packager if $changelog_packager;
    push @args, 'DATE' => $changelog_date if $changelog_date;
    $spec->add_changelog(@args);
}

sub __set_packager {
    my ($spec,$parent,$default_packager_policy,$packager) = @_;
    my $mainsec=$spec->get_section('package','');
    my $packager_tag=$mainsec->get_tag('Packager');
    if ($packager) {
	if ($packager eq 'auto') {
	    $packager=$spec->macros()->macro_subst('%{packager}');
	} elsif ($packager eq 'none') {
	    $mainsec->clear_tag('Packager');
	} elsif ($packager eq 'parent') {
	    die "--packager=parent but there is no parent\n" unless $parent;
	    my $candidate=$parent->get_section('package','')->get_tag('Packager');
	    if ($candidate) {
		$mainsec->set_tag('Packager',$candidate);
	    } else {
		$mainsec->clear_tag('Packager');
	    }
	} else {
	    $mainsec->set_tag('Packager',$packager);
	}
    } elsif (! $packager_tag and $default_packager_policy) {
	foreach my $policy (split(/,/,$default_packager_policy)) {
	    if ($policy eq 'parent') {
		next unless $parent;
		my $candidate=$parent->get_section('package','')->get_tag('Packager');
		if ($candidate) {
		    $mainsec->set_tag('Packager',$candidate);
		    last;
		}
	    } elsif ($policy eq 'auto') {
		my $candidate=$spec->macros()->macro_subst('%{packager}');
		if ($candidate) {
		    $mainsec->set_tag('Packager',$candidate);
		    last;
		}
	    } elsif ($policy =~ m'\@') {
		# default packager
		$mainsec->set_tag('Packager',$policy);
		last;
	    }
	}
    }
}


sub __serial_val {
    my ($epoch,$serial) =@_;
    if (defined $epoch) {
	if (defined $serial) {
	    return ($epoch > $serial ) ? $epoch:$serial;
	} else {
	    return $epoch;
	}
    } else {
	if (defined $serial) {
	    return $serial;
	} else {
	    return $epoch;
	}
    }
    return;
}

sub __set_serial {
    my ($spec,$parent,$serial,$epoch)=@_;
    my $specmain=$spec->get_section('package','');
    if ($parent and not defined $epoch and not defined $serial) {
	my $parentmain=$parent->get_section('package','');
	my $parentepoch = $parentmain->get_tag('Epoch');
	my $parentserial = $parentmain->get_tag('Serial');
	my $oldepoch = $specmain->get_tag('Epoch');
	my $oldserial = $specmain->get_tag('Serial');
	my $parentval= &__serial_val($parentepoch,$parentserial);
	my $oldval= &__serial_val($oldepoch,$oldserial);
	if (! defined $oldval || (defined $parentval and $parentval > $oldval)) {
	    if (defined $parentepoch) {
		$epoch=$parentepoch;
		$oldepoch=$parentepoch;
	    }
	    if (defined $parentserial) {
		$serial=$parentserial;
		$oldserial=$parentserial;
	    }
	    $oldval=$parentval;
	    print STDERR "note: picked up serial $parentserial from parent\n" if defined($parentserial) and $verbose;
	    print STDERR "note: picked up epoch $parentepoch from parent\n" if defined($parentepoch) and $verbose;
	}
	my $specver = $spec->get_section('package','')->get_tag('Version');
	my $parentver = $parent->get_section('package','')->get_tag('Version');
	my $vercmp = evrcmp($parentver,$specver);
	if ($vercmp > 0 and (! defined $oldval || (defined $parentval and $parentval >= $oldval))) {
	    if (defined($oldserial)) {
		$serial=++$oldserial;
		print STDERR "note: downgrade detected; increasing serial to $serial\n" if $verbose;
	    } elsif (defined($oldepoch)) {
		$epoch=++$oldepoch;
		print STDERR "note: downgrade detected; increasing epoch to $epoch\n" if $verbose;
	    } else {
		$serial=1;
		print STDERR "note: downgrade detected; increasing serial to $serial\n" if $verbose;
	    }
	}
    }

    if (defined $epoch) {
	if ($epoch eq 'none') {
	    $specmain->exclude(qr'^Epoch:');
	    undef $spec->macros()->{'epoch'};
	} else {
	    $specmain->set_tag('Epoch',$epoch);
	}
	$specmain->exclude(qr'^Serial:');
	undef $spec->macros()->{'serial'};
    }
    if (defined $serial) {
	if ($serial eq 'none') {
	    $specmain->exclude(qr'^Serial:');
	    undef $spec->macros()->{'serial'};
	} else {
	    $specmain->set_tag('Serial',$serial);
	}
	$specmain->exclude(qr'^Epoch:');
	undef $spec->macros()->{'epoch'};
    }
}

sub __increment_release() {
    my ($spec,$parent,$nextrel_policy,$forced_value,$force_transform_mode)=@_;
    my @old_evr=$spec->get_evr();
    my $old_vr="$old_evr[1]-$old_evr[2]";
    if (!defined ($forced_value) && $parent && !$force_transform_mode) {
	my @parent_evr=$parent->get_evr();
	my $vercmp = evrcmp($parent_evr[1],$old_evr[1]);
	if ($vercmp < 0) {
	    # upgrade;
	    $forced_value='alt1';
	    my $initrel_method=$INITREL_METHOD{$nextrel_policy};
	    $forced_value=&$initrel_method($spec,$parent) if $initrel_method;
	} elsif ($vercmp>0) {
	    # downgrade;
	    $forced_value=&__raw_increment_release($nextrel_policy, $parent_evr[2],$spec,$parent);
	    # TODO: check epoch and increment if nessecary
	} else {
	    $forced_value=&__raw_increment_release($nextrel_policy, $parent_evr[2],$spec,$parent);
	}
    } 
    my $mainsec=$spec->get_section('package','');
    my $body=$mainsec->get_body();
    for (my $i=0; $i<@$body; $i++) {
	if ($body->[$i]=~/^(release:\s+)(\S+)(.+)/si) {
	    my ($pre,$rel,$post)=($1,$2,$3);
	    $rel=$forced_value ? $forced_value : &__raw_increment_release($nextrel_policy, $rel,$spec,$parent);
	    $body->[$i]=$pre.$rel.$post;
	    $spec->macros()->{'release'}=$spec->macros()->macro_subst($rel);
	}
    }
    $mainsec->set_body($body);

    ## HACK to rename sources along with new release.
    my @new_evr=$spec->get_evr();
    my $new_vr="$new_evr[1]-$new_evr[2]";
    if ($spec->{SOURCERPM} and -d $spec->{SOURCEDIR} and $old_vr and $new_vr) {
	foreach my $file (glob "$spec->{SOURCEDIR}/*") {
	    next unless -f $file;
	    my $basename=basename($file);
	    my $copy=$basename;
	    $copy=~s/\Q$old_vr\E/$new_vr/;
	    if ($copy ne $basename) {
		system ('cp',$file,"$spec->{SOURCEDIR}/$copy")==0 or die "cp $file $spec->{SOURCEDIR}/$copy failed!";
	    }
	}

	my $prepsec=$spec->get_section('prep');
	#%setup -q -n %name-%version-%release
	if ($prepsec and $prepsec->match(qr'^\s*\%setup.+-n\s+.+-\%{?version}?-\%{?release}?')) {
	    my $repack=0;
	    foreach my $file (glob "$spec->{SOURCEDIR}/*") {
		next unless -f $file and $file=~/\.tar(\.(?:gz|bz2))?$/;
		my $arj='';
		if ($1) {
		    $arj='z' if ($1 eq '.gz');
		    $arj='j' if ($1 eq '.bz2');
		}
		my $tmpdir = tempdir( "tmp-unpack-source.XXXXXXXX", TMPDIR => 1, CLEANUP=>1);
		die "can't create temporary directory!" unless $tmpdir;
		#my $curdir=`pwd`;
		#chomp $curdir;
		#chdir $tmpdir;
		system('tar','x'.$arj.'f',$file,'-C',$tmpdir)==0 or die 'tar x'.$arj."f $file -C $tmpdir failed!";
		my ($dir)=glob "$tmpdir/*";
		if ( -d $dir) {
		    my $basename=basename($dir);
		    my $copy=$basename;
		    $copy=~s/\Q$old_vr\E/$new_vr/;
		    if ($copy ne $basename) {
			system ('mv','-f',$dir,"$tmpdir/$copy")==0 or die "mv $dir $tmpdir/$copy failed!";
			system ('tar','c'.$arj.'f',$file,'-C',$tmpdir,$copy)==0 or die 'tar c'.$arj."f $file -C $tmpdir $copy failed!";
			$repack=1;
		    }
		}
	    }
	    unless ($repack) {
		print STDERR "WARNING: found \%setup -q -n ...-\%version-\%release\n";
	    }
	}
    }
}

sub __raw_increment_release {
    my ($nextrel_policy,$rel,$spec,$parent)=@_;
    $nextrel_policy||= 'incr';
    my $nextrel_method=$NEXTREL_METHOD{$nextrel_policy};
    die "Unknown next release policy: $nextrel_policy" unless $nextrel_method;
    my $newrel = &{$nextrel_method}($rel,$spec,$parent);
    $newrel||= $rel.".1";
    return $newrel;
}

sub nextrel_incr {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~/^((?:alt)?)(\d+)/) {
        my $relprefix=$1;
        my $digitrel=$2;
        $newrel = $relprefix.++$digitrel;
    }
    return $newrel;
}

sub nextrel_incrfirst {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~/^((?:alt)?)(\d+)(.*)$/) {
        my $relprefix=$1;
        my $digitrel=$2; my $tail=$3;
        $newrel = $relprefix.++$digitrel.$tail;
    }
    return $newrel;
}

sub nextrel_incrsecond {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.(\d+)$//) {
	$newrel = $rel.".".($1+1);
    } else {
	$newrel = $rel.".1";
    }
    return $newrel;
}

sub nextrel_qa {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.qa(\d+)(?:\.(?:qa\d+|1))*$//) {
	$newrel = $rel.".qa".($1+1);
    } else {
	$newrel = $rel.".qa1";
    }
    return $newrel;
}

sub nextrel_nmuappend {
    my ($rel)=@_;
    return $rel.".1";
}

sub nextrel_nmuadd {
    my ($rel)=@_;
    my $newrel;
    if ($rel=~s/\.(\d+)$//) {
	$newrel = $rel.".".($1+1);
    } else {
	$newrel = $rel.".1";
    }
    return $newrel;
}

sub nextrel_none {
    my ($rel)=@_;
    return $rel;
}
